/* -*- Mode: C++ -*-
   vi:si:et:sw=4:sts=4:ts=4
 */
/*
  kerbsystray - A systray util to list kerberos tickets on Windows
  Copyright (C) 2007 Michele Baldessari <michele@pupazzo.org>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
*/

#include <windows.h>
#include "kerbutil.h"
#include "resource.h"

#define MAX_TMPSTR_LEN 127

static HWND hwnd = NULL;	
static HWND ticketlist = NULL;
static HWND advanceddialog = NULL;
static HINSTANCE hinst = NULL;
static HMENU systray_menu = NULL;
static HANDLE LsaHandle = NULL;
static ULONG AuthPackage;
static PKERB_QUERY_TKT_CACHE_RESPONSE pTicketCacheResponse = NULL;
static KerberosUtilClass *krb = NULL;


static BOOL 
addTrayIcon(HWND hwnd)
{
    BOOL res;
    NOTIFYICONDATA tnid;
    HICON hicon;

    tnid.cbSize = sizeof(NOTIFYICONDATA);
    tnid.hWnd = hwnd;
    tnid.uID = 1;	       /* unique within this systray use */
    tnid.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
    tnid.uCallbackMessage = WM_SYSTRAY;
	tnid.hIcon = hicon = LoadIcon(hinst, MAKEINTRESOURCE(KST_ICON));
	strncpy_s(tnid.szTip, strlen(APPNAME) + 1, APPNAME, strlen(APPNAME) + 1);
    res = Shell_NotifyIcon(NIM_ADD, &tnid);

    if (hicon) 
		DestroyIcon(hicon);
    return res;
}

void 
quit(int ret)
{
	NOTIFYICONDATA tnid;

	tnid.cbSize = sizeof(NOTIFYICONDATA);
	tnid.hWnd = hwnd;
	tnid.uID = 1;
	Shell_NotifyIcon(NIM_DELETE, &tnid);
	DestroyMenu(systray_menu);
	delete krb;
	exit(ret);
}

void 
updateTicketList(void)
{
	KERB_TICKET_CACHE_INFO ticket;
	char str[MAX_TMPSTR_LEN];
	unsigned int i;
	size_t converted_chars, orig_size;

	if (ticketlist == NULL || !krb->isInitialized())
		return;

	krb->refreshKerberosTicketCache();
	SendDlgItemMessage(ticketlist, TICKETLIST_LISTBOX, 
			LB_RESETCONTENT, 0, 0);
	for (i = 0; i < krb->getTicketCacheCount(); i++) {
		if (!krb->getTicketFromCache(i, &ticket))
			return;
		memset(str, 0, MAX_TMPSTR_LEN);
		orig_size = ticket.ServerName.Length / sizeof(wchar_t);
		wcstombs_s(&converted_chars, str, MAX_TMPSTR_LEN, ticket.ServerName.Buffer, orig_size);
		SendDlgItemMessage(ticketlist, TICKETLIST_LISTBOX, LB_ADDSTRING, 0, (LPARAM)str);
	}
}

static int
CALLBACK procAdvancedTicket(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
      case WM_INITDIALOG:
	  case WM_CREATE:
		//MessageBox(NULL, "TEST", "Error", MB_OK | MB_ICONERROR);


	    RECT rs, rd;
	    HWND hw;
	    hw = GetDesktopWindow();
	    if (GetWindowRect(hw, &rs) && GetWindowRect(hwnd, &rd))
			MoveWindow(hwnd,
			   (rs.right + rs.left + rd.left - rd.right) / 2,
			   (rs.bottom + rs.top + rd.top - rd.bottom) / 2,
			   rd.right - rd.left, rd.bottom - rd.top, TRUE);
		break;
      case WM_COMMAND:
		switch (LOWORD(wParam)) {
			case IDOK:
				advanceddialog = NULL;
				DestroyWindow(hwnd);
				break;
		}
		break;
	  case WM_CLOSE:
		advanceddialog = NULL;
		DestroyWindow(hwnd);
		break;
	}
	return 0;
}

void displayAdvancedTicket(unsigned int ticket_nr)
{
	PKERB_EXTERNAL_TICKET tkt = NULL;
	char *tmp;

	if (!krb->getMSCacheTicketFromCacheInfo(ticket_nr, &tkt)) {
		MessageBox(NULL, "Display Advanced Ticket", "Error", MB_OK | MB_ICONERROR);
		return;
	}
	if (advanceddialog == NULL) {
		advanceddialog = CreateDialog(hinst, MAKEINTRESOURCE(KSTDLG_TICKETADVANCED), 
			hwnd, procAdvancedTicket);
		if (advanceddialog == NULL) {
			MessageBox(NULL, "Error creating advanced dialog", "Error", MB_OK | MB_ICONERROR);
			return;
		}
		ShowWindow(advanceddialog, SW_SHOWNORMAL);
		SetForegroundWindow(advanceddialog);
		SetWindowPos(advanceddialog, HWND_TOP, 0, 0, 0, 0,
			SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW);
		
		tmp = char_to_hex((char *)tkt->EncodedTicket);
		SetDlgItemText(advanceddialog, ADVANCED_EC_TICKETDUMP, tmp);
		free(tmp);

		tmp = char_to_hex((char *)tkt->SessionKey.Value);
		SetDlgItemText(advanceddialog, ADVANCED_EC_SESSIONKEYDUMP, tmp);
		free(tmp);
			
	}
}

static int 
CALLBACK procTicketList(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	LRESULT selected, count;
	char str[MAX_TMPSTR_LEN];
	size_t orig_size, converted_chars;
	FILETIME tmp_filetime;
	SYSTEMTIME tmp_systime;
	KERB_TICKET_CACHE_INFO ticket;
	PKERB_EXTERNAL_TICKET tkt = NULL;

    switch (msg) {
      case WM_INITDIALOG:
        /* centre the window */
	    RECT rs, rd;
	    HWND hw;
	    hw = GetDesktopWindow();
	    if (GetWindowRect(hw, &rs) && GetWindowRect(hwnd, &rd))
			MoveWindow(hwnd,
			   (rs.right + rs.left + rd.left - rd.right) / 2,
			   (rs.bottom + rs.top + rd.top - rd.bottom) / 2,
			   rd.right - rd.left, rd.bottom - rd.top, TRUE);
		break;
      case WM_COMMAND:
		switch (LOWORD(wParam)) {
			case TICKETLIST_BT_ADVANCED:
				count = SendDlgItemMessage(ticketlist, TICKETLIST_LISTBOX, LB_GETCOUNT, 0, 0);
				if (count < 1)
					break;
				selected = SendDlgItemMessage(ticketlist, TICKETLIST_LISTBOX, LB_GETCURSEL, 0, 0);
				if (!krb->isInitialized() || selected < 0)
					break;
				displayAdvancedTicket((unsigned int)selected);
				break;
			case TICKETLIST_BT_CLOSE:
				ticketlist = NULL;
				DestroyWindow(hwnd);
				return 0;
			case TICKETLIST_LISTBOX:
				if ((HIWORD(wParam) == LBN_SELCHANGE)) {
					count = SendDlgItemMessage(ticketlist, TICKETLIST_LISTBOX, LB_GETCOUNT, 0, 0);
					if (count < 1)
						break;
					selected = SendDlgItemMessage(ticketlist, TICKETLIST_LISTBOX, LB_GETCURSEL, 0, 0);
					if (!krb->isInitialized() || selected < 0)
						break;
					if (!krb->getTicketFromCache((unsigned int)selected, &ticket))
						break;

					if (!krb->getMSCacheTicketFromCacheInfo((unsigned int)selected, &tkt))
						tkt = NULL;
					orig_size = ticket.ServerName.Length / sizeof(wchar_t);
					wcstombs_s(&converted_chars, str, MAX_TMPSTR_LEN, ticket.ServerName.Buffer, orig_size);
					SetDlgItemText(ticketlist, TICKETLIST_EC_SERVERNAME, str);
					orig_size = ticket.RealmName.Length / sizeof(wchar_t);
					wcstombs_s(&converted_chars, str, MAX_TMPSTR_LEN, ticket.RealmName.Buffer, orig_size);
					SetDlgItemText(ticketlist, TICKETLIST_EC_REALMNAME, str);
					memset(str, 0, MAX_TMPSTR_LEN);
					tmp_filetime.dwLowDateTime = ticket.StartTime.LowPart;
					tmp_filetime.dwHighDateTime = ticket.StartTime.HighPart;
					FileTimeToSystemTime(&tmp_filetime, &tmp_systime);
					sprintf_s(str, MAX_TMPSTR_LEN, "%4d-%02d-%02d %02d:%02d,%02d", tmp_systime.wYear,
						tmp_systime.wMonth, tmp_systime.wDay, tmp_systime.wHour, tmp_systime.wMinute,
						tmp_systime.wSecond);
					SetDlgItemText(ticketlist, TICKETLIST_EC_STARTTIME, str);

					tmp_filetime.dwLowDateTime = ticket.EndTime.LowPart;
					tmp_filetime.dwHighDateTime = ticket.EndTime.HighPart;
					FileTimeToSystemTime(&tmp_filetime, &tmp_systime);
					sprintf_s(str, MAX_TMPSTR_LEN, "%4d-%02d-%02d %02d:%02d,%02d", tmp_systime.wYear,
						tmp_systime.wMonth, tmp_systime.wDay, tmp_systime.wHour, tmp_systime.wMinute,
						tmp_systime.wSecond);
					SetDlgItemText(ticketlist, TICKETLIST_EC_ENDTIME, str);

					tmp_filetime.dwLowDateTime = ticket.RenewTime.LowPart;
					tmp_filetime.dwHighDateTime = ticket.RenewTime.HighPart;
					FileTimeToSystemTime(&tmp_filetime, &tmp_systime);
					sprintf_s(str, MAX_TMPSTR_LEN, "%4d-%02d-%02d %02d:%02d,%02d", tmp_systime.wYear,
						tmp_systime.wMonth, tmp_systime.wDay, tmp_systime.wHour, tmp_systime.wMinute,
						tmp_systime.wSecond);
					SetDlgItemText(ticketlist, TICKETLIST_EC_RENEWTIME, str);
					/* Uncheck them all */
					SendDlgItemMessage(ticketlist, TICKETLIST_CB_FORWARDED, BM_SETCHECK, BST_UNCHECKED, 0);
					SendDlgItemMessage(ticketlist, TICKETLIST_CB_FORWARDABLE, BM_SETCHECK, BST_UNCHECKED, 0);
					SendDlgItemMessage(ticketlist, TICKETLIST_CB_PREAUTHENTICATED, BM_SETCHECK, BST_UNCHECKED, 0);
					SendDlgItemMessage(ticketlist, TICKETLIST_CB_RENEWABLE, BM_SETCHECK, BST_UNCHECKED, 0);
					SendDlgItemMessage(ticketlist, TICKETLIST_CB_INITIAL, BM_SETCHECK, BST_UNCHECKED, 0);
					SendDlgItemMessage(ticketlist, TICKETLIST_CB_POSTDATED, BM_SETCHECK, BST_UNCHECKED, 0);				
					SendDlgItemMessage(ticketlist, TICKETLIST_CB_HWAUTHENTICATED, BM_SETCHECK, BST_UNCHECKED, 0);
					SendDlgItemMessage(ticketlist, TICKETLIST_CB_OKASDELEGATE, BM_SETCHECK, BST_UNCHECKED, 0);
					SendDlgItemMessage(ticketlist, TICKETLIST_CB_PROXIABLE, BM_SETCHECK, BST_UNCHECKED, 0);
					SendDlgItemMessage(ticketlist, TICKETLIST_CB_PROXY, BM_SETCHECK, BST_UNCHECKED, 0);
					SendDlgItemMessage(ticketlist, TICKETLIST_CB_MAYPOSTDATE, BM_SETCHECK, BST_UNCHECKED, 0);

					if (ticket.TicketFlags & KERB_TICKET_FLAGS_forwardable)
						SendDlgItemMessage(ticketlist, TICKETLIST_CB_FORWARDABLE, BM_SETCHECK, BST_CHECKED, 0);
					if (ticket.TicketFlags & KERB_TICKET_FLAGS_forwarded)
						SendDlgItemMessage(ticketlist, TICKETLIST_CB_FORWARDED, BM_SETCHECK, BST_CHECKED, 0);
					if (ticket.TicketFlags & KERB_TICKET_FLAGS_pre_authent)
						SendDlgItemMessage(ticketlist, TICKETLIST_CB_PREAUTHENTICATED, BM_SETCHECK, BST_CHECKED, 0);
					if (ticket.TicketFlags & KERB_TICKET_FLAGS_renewable)
						SendDlgItemMessage(ticketlist, TICKETLIST_CB_RENEWABLE, BM_SETCHECK, BST_CHECKED, 0);
					if (ticket.TicketFlags & KERB_TICKET_FLAGS_initial)
						SendDlgItemMessage(ticketlist, TICKETLIST_CB_INITIAL, BM_SETCHECK, BST_CHECKED, 0);
					if (ticket.TicketFlags & KERB_TICKET_FLAGS_hw_authent)
						SendDlgItemMessage(ticketlist, TICKETLIST_CB_HWAUTHENTICATED, BM_SETCHECK, BST_CHECKED, 0);
					if (ticket.TicketFlags & KERB_TICKET_FLAGS_ok_as_delegate)
						SendDlgItemMessage(ticketlist, TICKETLIST_CB_OKASDELEGATE, BM_SETCHECK, BST_CHECKED, 0);
					if (ticket.TicketFlags & KERB_TICKET_FLAGS_proxiable)
						SendDlgItemMessage(ticketlist, TICKETLIST_CB_PROXIABLE, BM_SETCHECK, BST_CHECKED, 0);
					if (ticket.TicketFlags & KERB_TICKET_FLAGS_proxy)
						SendDlgItemMessage(ticketlist, TICKETLIST_CB_PROXY, BM_SETCHECK, BST_CHECKED, 0);
					if (ticket.TicketFlags & KERB_TICKET_FLAGS_may_postdate)
						SendDlgItemMessage(ticketlist, TICKETLIST_CB_MAYPOSTDATE, BM_SETCHECK, BST_CHECKED, 0);
					if (ticket.TicketFlags & KERB_TICKET_FLAGS_postdated)
						SendDlgItemMessage(ticketlist, TICKETLIST_CB_POSTDATED, BM_SETCHECK, BST_CHECKED, 0);
					memset(str, 0, MAX_TMPSTR_LEN);
	
					switch (ticket.EncryptionType) {
						case KERB_ETYPE_DES_CBC_MD5_NT:
							sprintf_s(str, MAX_TMPSTR_LEN, "DES-CBC-MD5");
							break;
						case KERB_ETYPE_RC4_HMAC_NT:
							sprintf_s(str, MAX_TMPSTR_LEN, "RC4-HMAC");
							break;
						case KERB_ETYPE_RC4_HMAC_NT_EXP:
							sprintf_s(str, MAX_TMPSTR_LEN, "RC4-HMAC-XP");
							break;
						default:
							sprintf_s(str, MAX_TMPSTR_LEN, "%d", ticket.EncryptionType);
							break;
					}
					SetDlgItemText(ticketlist, TICKETLIST_EC_TKTENCTYPE, str);
					if (tkt == NULL)
						break;
					
					// We were able to retrieve the whole ticket
					tmp_filetime.dwLowDateTime = tkt->KeyExpirationTime.LowPart;
					tmp_filetime.dwHighDateTime = tkt->KeyExpirationTime.HighPart;
					FileTimeToSystemTime(&tmp_filetime, &tmp_systime);
					if (tmp_filetime.dwHighDateTime != 0L &&
						tmp_filetime.dwLowDateTime != 0L)
						sprintf_s(str, MAX_TMPSTR_LEN, "%4d-%02d-%02d %02d:%02d,%02d", tmp_systime.wYear,
							tmp_systime.wMonth, tmp_systime.wDay, tmp_systime.wHour, tmp_systime.wMinute,
							tmp_systime.wSecond);
					else
						sprintf_s(str, MAX_TMPSTR_LEN, "No expiration");
					SetDlgItemText(ticketlist, TICKETLIST_EC_KEYEXPIRATION, str);

					tmp_filetime.dwLowDateTime = tkt->TimeSkew.LowPart;
					tmp_filetime.dwHighDateTime = tkt->TimeSkew.HighPart;
					FileTimeToSystemTime(&tmp_filetime, &tmp_systime);
					if (tmp_filetime.dwHighDateTime != 0L &&
						tmp_filetime.dwLowDateTime != 0L)
						sprintf_s(str, MAX_TMPSTR_LEN, "%4d-%02d-%02d %02d:%02d,%02d", tmp_systime.wYear,
							tmp_systime.wMonth, tmp_systime.wDay, tmp_systime.wHour, tmp_systime.wMinute,
							tmp_systime.wSecond);
					else
						sprintf_s(str, MAX_TMPSTR_LEN, "No time skew");
					SetDlgItemText(ticketlist, TICKETLIST_EC_TIMESKEW, str);

					orig_size = tkt->TargetDomainName.Length / sizeof(wchar_t);
					wcstombs_s(&converted_chars, str, MAX_TMPSTR_LEN, tkt->TargetDomainName.Buffer, orig_size);
					SetDlgItemText(ticketlist, TICKETLIST_EC_TARGETDOMAIN, str);
					
					// FIXME: Check all Names[0]
					orig_size = tkt->ClientName->Names[0].Length / sizeof(wchar_t);
					wcstombs_s(&converted_chars, str, MAX_TMPSTR_LEN, tkt->ClientName->Names[0].Buffer, orig_size);
					sprintf_s(str, MAX_TMPSTR_LEN, "%s - %d", str, tkt->ClientName->NameType);
					SetDlgItemText(ticketlist, TICKETLIST_EC_CLIENTNAME, str);

					LsaFreeReturnBuffer(tkt);
					tkt = NULL;
				break;
			}
		}
		break;
      case WM_CLOSE:
		ticketlist = NULL;
		DestroyWindow(hwnd);
		break;
	  default:
		break;
    }
    return 0;
}

static LRESULT CALLBACK 
procWnd(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	int ret;
    static int menuinprogress;
    static UINT msgTaskbarCreated = 0;

    switch (message) {
      case WM_CREATE:
        msgTaskbarCreated = RegisterWindowMessage("TaskbarCreated");
        break;
      case WM_COMMAND:
      case WM_SYSCOMMAND:
		switch (wParam & ~0xF) {       /* low 4 bits reserved to Windows */
		  case IDM_CLOSE:
			quit(0);
			break;
		  case IDM_ABOUT:
			MessageBox(NULL, "kerbsystray \n\nCopyright (c) 2007 - Michele Baldessari <michele@pupazzo.org>\n Released " \
				"under the terms of the GPL License",
				   "About", MB_OK | MB_ICONINFORMATION);		
		    break;
		  case IDM_LIST:
			if (ticketlist == NULL) {
				ticketlist = CreateDialog(hinst, MAKEINTRESOURCE(KSTDLG_TICKETLIST), 
					hwnd, procTicketList);
				if (ticketlist == NULL) {
					MessageBox(NULL, "Error creating ticketlist dialog", "Error", MB_OK | MB_ICONERROR);
					break;
				}
				ShowWindow(ticketlist, SW_SHOWNORMAL);
				SetForegroundWindow(ticketlist);
				SetWindowPos(ticketlist, HWND_TOP, 0, 0, 0, 0,
					SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW);
				updateTicketList();
			}
			break;
		  case IDM_PURGE:
			  if (krb->purgeAllTickets())
				  MessageBox(hwnd, "All tickets purged successfully", "Purge", MB_OK | MB_ICONINFORMATION);
			  else
				  MessageBox(hwnd, "Error purging tickets", "Purge", MB_OK | MB_ICONERROR);
			  break;
		}
		break;
      default:
        if (message==msgTaskbarCreated) {
            /*
		    * Explorer has been restarted, so the tray icon will
			* have been lost.
			*/
			addTrayIcon(hwnd);
        }
        break;
      case WM_SYSTRAY:
		if (lParam == WM_RBUTTONUP) {
			POINT cursorpos;
			GetCursorPos(&cursorpos);
			PostMessage(hwnd, WM_SYSTRAY2, cursorpos.x, cursorpos.y);
		} else if (lParam == WM_LBUTTONDBLCLK) {
			/* Run the default menu item. */
			UINT menuitem = GetMenuDefaultItem(systray_menu, FALSE, 0);
			if (menuitem != -1)
				PostMessage(hwnd, WM_COMMAND, menuitem, 0);
		}
		break;
      case WM_SYSTRAY2:
			SetForegroundWindow(hwnd);
			ret = TrackPopupMenu(systray_menu,
				 TPM_RIGHTALIGN | TPM_BOTTOMALIGN |
				 TPM_RIGHTBUTTON, (int)wParam, (int)lParam, 0, hwnd, NULL);
		
		
	} /* switch (message) */
    return DefWindowProc(hwnd, message, wParam, lParam);
}

int WINAPI 
WinMain(HINSTANCE inst, 
		HINSTANCE prev,
		LPSTR cmdline, 
		int show)
{
	WNDCLASS wndclass;
    MSG msg;

	krb = new KerberosUtilClass();
	if (!krb->isInitialized()) {
		MessageBox(NULL, "Error initializing Kerberos class",
				   "Error", MB_OK | MB_ICONINFORMATION);
		return -1;
	}

	hinst = inst;
	wndclass.style = 0;
	wndclass.lpfnWndProc = procWnd;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = inst;
	wndclass.hIcon = NULL;
	wndclass.hCursor = LoadCursor(NULL, IDC_IBEAM);
	wndclass.hbrBackground = NULL;
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = APPNAME;
	RegisterClass(&wndclass);

    hwnd = CreateWindow(APPNAME, APPNAME,
			WS_OVERLAPPEDWINDOW | WS_VSCROLL,
			CW_USEDEFAULT, CW_USEDEFAULT,
			100, 100, NULL, NULL, hinst, NULL);
	addTrayIcon(hwnd);
	systray_menu = CreatePopupMenu();
	AppendMenu(systray_menu, MF_ENABLED, IDM_LIST,  "&List Tickets");
	AppendMenu(systray_menu, MF_ENABLED, IDM_PURGE, "&Purge Tickets");
    AppendMenu(systray_menu, MF_SEPARATOR, 0, 0);
	AppendMenu(systray_menu, MF_ENABLED, IDM_ABOUT,  "&About");
    AppendMenu(systray_menu, MF_ENABLED, IDM_CLOSE,  "E&xit");

    /*
     * Main message loop.
     */
    while (GetMessage(&msg, NULL, 0, 0) == 1) {
		if (!(IsWindow(ticketlist) && IsDialogMessage(ticketlist, &msg))) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
    }

	quit(0);
	return 0; // Compiler is too dumb anyway
}
